from typing import NamedTuple, Sequence
import asyncio
from sqlmodel import Session, select
from haskellian import either as E
from kv.api import KV

from moveread.dfy.types import Game, Pairing, SheetModel
from moveread.pipelines.dfy import Input

def pairing_display(pairing: Pairing):
  pair = pairing.root
  if pair.tag == 'unpaired':
    return 'Unpaired!?'
  
  s = f'{pair.white} - {pair.black}'
  if pair.result is not None:
    s += f' {pair.result}'
  return s


def title(pairing: Pairing, tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId} {group}/{round}/{board} {pairing_display(pairing)}'

class NewInput(NamedTuple):
  input: Input
  """Pipeline input"""
  urls: Sequence[str]
  """Blob URLs to be copied"""

def new_input(game: Game, *, model: str):
  gid = game.gameId()
  urls = [img.url for img in game.imgs]
  endpoint = f'/v1/models/{gid["tournId"]}-{gid["group"]}:predict'
  task = Input(gameId=gid, model=model, imgs=urls, title=title(game.pairing, **gid), serving_endpoint=endpoint)
  return NewInput(input=task, urls=urls)


def sheet_model(session: Session, tournId: str) -> str | None:
  stmt = select(SheetModel).where(SheetModel.tournId == tournId)
  obj = session.exec(stmt).first()
  return obj and obj.model


async def copy_images(
  urls: Sequence[str], *,
  online_images: KV[bytes], pipeline_images: KV[bytes]
):
  tasks = [online_images.copy(url, pipeline_images, url).run() for url in urls]
  results = await asyncio.gather(*tasks)
  return E.sequence(results)



from time import time
from fastapi import Request

def from_lichess(req: Request):
  return 'lichess' in req.headers.get('User-Agent', '')

def clientId(req: Request) -> str | None:
  """Client's Host, fallbacked to User-Agent"""
  return req.client.host if req.client else req.headers.get('User-Agent')

class Throttler:
  def __init__(self, max_entries: int, ttl_secs: float):
    self.last_accesses: dict[str, float] = {}
    self.max_entries = max_entries
    self.ttl_secs = ttl_secs

  def insert(self, client: str, access_time: float):
    if len(self.last_accesses) >= self.max_entries:
      older_client = min(self.last_accesses.keys(), key=self.last_accesses.__getitem__)
      del self.last_accesses[older_client]
    self.last_accesses[client] = access_time

  def throttle(self, client: str) -> bool:
    now = time()
    if not client in self.last_accesses:
      self.insert(client, now)
      return False
    elif now - self.last_accesses[client] > self.ttl_secs:
      self.insert(client, now)
      return False
    else:
      delta = now - self.last_accesses[client]
      print(f'Last accessed {delta:.2f} secs ago by {client}')
      return True


from typing import Sequence
from datetime import datetime, timedelta
from sqlmodel import select, Session
import chess_pairings as cp
from moveread.dfy.types import Group, Pairings, Game, Pairing, Tournament
from moveread.dfy.lib import queries
from haskellian import Either, Left, Right
from dslog import Logger

def current_pairings(session: Session, now: datetime):
  tomorrow = now + timedelta(days=1)
  yesterday = now - timedelta(days=1)
  stmt = select(Tournament, Pairings) \
    .where(Tournament.start_date < tomorrow, yesterday < Tournament.end_date) \
    .join(Pairings, Tournament.tournId == Pairings.tournId) # type: ignore
  return session.exec(stmt).all()

def update_rounds(session: Session, tnmt: Tournament, group: str, rounds: Sequence[str]) -> Either[str, str]:
  """Update the group's `rounds` list"""
  grp = session.get(Group, (tnmt.tournId, group))
  if grp is None:
    return Left(f'Group "{tnmt.tournId}/{group}" not found')
  elif grp.rounds != rounds:
    og_rounds = [*grp.rounds]
    grp.rounds = rounds
    session.add(grp)
    session.commit()
    return Right(f'Updated group "{tnmt.tournId}/{group}". Rounds {og_rounds} -> {rounds}')
  else:
    return Right(f'Group "{tnmt.tournId}/{group}" already has rounds {rounds}')
  

def update_pairings(session: Session, *, tournId: str, group: str, pairings: cp.GroupPairings):
    """Insert/update pairings"""
    group_games = session.exec(queries.select_group_games(tournId, group)).all()
    games_idx = cp.GamesMapping[Game].from_pairs([(g.gameId(), g) for g in group_games])
    added: list[cp.GameId] = []
    updated: list[cp.GameId] = []

    for round, rnd_pairings in pairings.items():
      for board, pair in rnd_pairings.items():
        gid = cp.gameId(tournId, group, round, board)
        if gid in games_idx:
          game = games_idx[gid]
          if game.pairing.root != pair:
            game.pairing = Pairing(pair)
            updated.append(gid)
        else:
          game = Game(tournId=tournId, group=group, round=round, board=board, index=int(board)-1, pairing=Pairing(pair))
          added.append(gid)
        
        session.add(game)
    
    session.commit()
    return added, updated


async def update_current_pairings(session: Session, *, now: datetime, logger: Logger):
  """Updates pairings for all ongoing tournaments"""
  pairing_sources = current_pairings(session, now)
  current_tnmts = [tnmt.tournId for tnmt, _ in pairing_sources]
  logger(f'Updating current tournaments: {", ".join(current_tnmts)}', level='DEBUG')
  for tnmt, src in pairing_sources:
    e = await cp.scrape_pairings(src.pairings)
    if e.tag == 'left':
      logger(f'Error fetching pairings for "{tnmt.tournId}/{src.group}"', e.value, level='ERROR')
      continue
    pairings = e.value
    rounds = [str(i+1) for i in range(len(pairings))]
    r = update_rounds(session, tnmt, src.group, rounds)
    logger(r.value, level='DEBUG' if r.tag == 'right' else 'ERROR')
    added, updated = update_pairings(session, tournId=tnmt.tournId, group=src.group, pairings=pairings)
    logger(f'Updated pairings for "{tnmt.tournId}/{src.group}", added {len(added)} and updated {len(updated)} games', level='DEBUG')